package cn.com.qmth.examcloud.starters.face.verify.model.baidu;

import cn.com.qmth.examcloud.starters.face.verify.FaceVerifyProperties;
import cn.com.qmth.examcloud.starters.face.verify.common.Constants;
import cn.com.qmth.examcloud.starters.face.verify.common.FaceVerifyException;
import cn.com.qmth.examcloud.starters.face.verify.common.HttpClientBuilder;
import cn.com.qmth.examcloud.starters.face.verify.common.JsonHelper;
import cn.com.qmth.examcloud.starters.face.verify.model.FaceResult;
import okhttp3.*;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.List;

public class BaiduApiHelper {

    private final static Logger log = LoggerFactory.getLogger(BaiduApiHelper.class);

    public static BaiduSession getAccessToken(String apiKey, String apiSecret) {
        if (StringUtils.isBlank(apiKey)) {
            throw new IllegalArgumentException("[BAIDU] apiKey must be not empty.");
        }

        if (StringUtils.isBlank(apiSecret)) {
            throw new IllegalArgumentException("[BAIDU] apiSecret must be not empty.");
        }

        String requestUrl = String.format("%s?grant_type=client_credentials&client_id=%s&client_secret=%s", Constants.BAIDU_TOKEN_URL, apiKey, apiSecret);
        Request.Builder request = new Request.Builder().url(requestUrl).get();

        try (Response response = HttpClientBuilder.getClient().newCall(request.build()).execute();
             ResponseBody body = response.body();) {
            String bodyStr = body.string();

            if (response.isSuccessful()) {
                BaiduSession session = new JsonHelper().parseJson(bodyStr, BaiduSession.class);
                if (session != null) {
                    return session;
                }
            }

            log.warn("[BAIDU] url:{} response:{} body:{}", requestUrl, response.code(), bodyStr);
            throw new FaceVerifyException("[BAIDU] 获取AccessToken失败！");
        } catch (IOException e) {
            log.error(e.getMessage(), e);
            throw new FaceVerifyException("[BAIDU] 获取AccessToken异常！");
        }
    }

    public static FaceResult faceVerify(FaceVerifyProperties properties, String jsonParams) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        BaiduSession session = BaiduApiHelper.getAccessToken(properties.getBaiduKey(), properties.getBaiduSecret());
        String accessToken = session.getAccess_token();

        String requestUrl = String.format("%s?access_token=%s", Constants.BAIDU_FACE_VERIFY_URL, accessToken);
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceDetectResult(response);
    }

    public static FaceResult faceVerifyUseLocalApi(FaceVerifyProperties properties, String jsonParams) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        String requestUrl = String.format("%s?appid=%s", properties.getBaiduLocalUrlPrefix() + Constants.BAIDU_FACE_VERIFY_LOCAL_URL, properties.getBaiduLocalAppId());
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceDetectResult(response);
    }

    public static FaceResult faceDetect(FaceVerifyProperties properties, String jsonParams) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        BaiduSession session = BaiduApiHelper.getAccessToken(properties.getBaiduKey(), properties.getBaiduSecret());
        String accessToken = session.getAccess_token();

        String requestUrl = String.format("%s?access_token=%s", Constants.BAIDU_FACE_DETECT_URL, accessToken);
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceDetectResult(response);
    }

    public static FaceResult faceDetectUseLocalApi(FaceVerifyProperties properties, String jsonParams) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        String requestUrl = String.format("%s?appid=%s", properties.getBaiduLocalUrlPrefix() + Constants.BAIDU_FACE_DETECT_LOCAL_URL, properties.getBaiduLocalAppId());
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceDetectResult(response);
    }

    public static FaceResult faceCompare(FaceVerifyProperties properties, String jsonParams, Double expectFaceCompareScore) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        BaiduSession session = BaiduApiHelper.getAccessToken(properties.getBaiduKey(), properties.getBaiduSecret());
        String accessToken = session.getAccess_token();

        String requestUrl = String.format("%s?access_token=%s", Constants.BAIDU_FACE_COMPARE_URL, accessToken);
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceCompareResult(response, expectFaceCompareScore);
    }

    public static FaceResult faceCompareUseLocalApi(FaceVerifyProperties properties, String jsonParams, Double expectFaceCompareScore) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        String requestUrl = String.format("%s?appid=%s", properties.getBaiduLocalUrlPrefix() + Constants.BAIDU_FACE_COMPARE_LOCAL_URL, properties.getBaiduLocalAppId());
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceCompareResult(response, expectFaceCompareScore);
    }

    private static FaceResult parseFaceDetectResult(BaiduResponse response) {
        FaceResult result = new FaceResult();
        result.setPass(false);
        result.setApiLimit(false);
        result.setScore(0d);
        result.setFaceNum(0);

        if (response.getError_code() != 0) {
            result.setError(response.getError_code() + " - " + response.getError_msg());

            // 请求并发超限(约10个/秒)
            if (response.getError_code() == 18) {
                result.setApiLimit(true);
            }
            return result;
        }

        BaiduResult data = response.getResult();
        if (data != null) {
            if (data.getThresholds() != null) {
                result.setThresholds(new JsonHelper().toJson(data.getThresholds()));
            }

            double faceLiveness = 0d;
            List<BaiduFace> faces = data.getFace_list();
            if (data.getFace_liveness() != null) {
                faceLiveness = data.getFace_liveness();
            } else {
                if (faces != null && !faces.isEmpty()) {
                    BaiduFace face = faces.get(0);
                    if (face.getLiveness() != null && face.getLiveness().getLivemapscore() != null) {
                        faceLiveness = face.getLiveness().getLivemapscore();
                    }
                }
            }
            result.setScore(faceLiveness);

            if (data.getFace_num() != null) {
                result.setFaceNum(data.getFace_num());
            } else {
                result.setFaceNum(faces != null ? faces.size() : 0);
            }

            // 活体分数值，推荐阈值0.393241
            if (faceLiveness >= 0.39d) {
                result.setPass(true);
            }
        }

        return result;
    }

    private static FaceResult parseFaceCompareResult(BaiduResponse response, Double expectFaceCompareScore) {
        FaceResult result = new FaceResult();
        result.setPass(false);
        result.setApiLimit(false);
        result.setScore(0d);

        if (response.getError_code() != 0) {
            result.setError(response.getError_code() + " - " + response.getError_msg());

            // 请求并发超限(约10个/秒)
            if (response.getError_code() == 18) {
                result.setApiLimit(true);
            }
            return result;
        }

        BaiduResult data = response.getResult();
        if (data != null) {
            result.setScore(data.getScore() != null ? data.getScore() : 0d);

            if (data.getScore() != null && data.getScore() >= expectFaceCompareScore) {
                result.setPass(true);
            }
        }

        return result;
    }

    private static BaiduResponse callApi(String requestUrl, RequestBody formBody) {
        long start = System.currentTimeMillis();
        Request.Builder request = new Request.Builder().url(requestUrl).post(formBody);
        try (Response response = HttpClientBuilder.getClient().newCall(request.build()).execute();
             ResponseBody body = response.body();) {
            String bodyStr = body.string();

            long cost = System.currentTimeMillis() - start;
            // System.out.println("response:" + response.code() + " cost:" + cost + "ms body:" + bodyStr);

            if (response.isSuccessful()) {
                BaiduResponse result = new JsonHelper().parseJson(bodyStr, BaiduResponse.class);
                if (result != null) {
                    return result;
                }
            }

            log.warn("[BAIDU] url:{} response:{} cost:{}ms body:{}", requestUrl, response.code(), cost, bodyStr);
        } catch (IOException e) {
            log.error(e.getMessage(), e);
        }

        throw new FaceVerifyException("[BAIDU] 人脸接口调用异常！");
    }

}
